/**************************************************************
 SVPFlow1 - motion vectors search for SmoothVideo Project (SVP)
 Copyright(c)2011 Chainik - http://www.svp-team.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
 http://www.gnu.org/copyleft/gpl.html
*****************************************************************/

#pragma once

#include "common.h"
#include <string>
#include <vector>
#include <map>
#include <list>

class RemoteControl
{
public:
    struct Message
    {
        int id;
        std::vector<int> args;
        std::vector<uint8_t> raw;

        int toByte(uint8_t *buf) const;
        int fromByte(const uint8_t* buf, int len);
    };

    typedef std::list<Message> MessageList;

public:
    static void create(uint64_t client);
    static void remove(uint64_t client);

    static MessageList exchange(uint64_t client, const MessageList& list = MessageList());

protected:
    RemoteControl();
    ~RemoteControl();
    bool tryToConnect();
    MessageList exchange_p(uint64_t client, const MessageList& out);

protected:
#ifdef Q_OS_WIN
    HANDLE hPipe;
#else
    int hPipe;
#endif
    std::chrono::steady_clock::time_point lastConnectTry;
    std::map<uint64_t,MessageList> messages;
};

